/*
 * This file implements the software-based I2C for accessing I2C
 * devices via the sensor ports.
 */

/*
 * original:lejos-osek/src/nxtvm/platform/nxt/i2c.c rev.1.1.1.1
 * modified history.
 * 05/24/2012: Modifier Takuya Ishikawa
 *           : include string.h
 * 05/23/2011: Modifier Takuya Ishikawa
 * 05/25/2010: Modifier Takuya Ishikawa
 *           : Moved i2c timer initialize to tSensorDriverBody.c
 *           : Moved i2c ISR entry to tSensorDriverBody.c
 *           : Modified port number to component attribute
 *           : Changed interface name to entry function name with TECS naming rule
 *           : Modified some variables to component variables
 *           : Modified some constants to component attributes
 *           : Modified data type to ITRON spec.
 * 04/15/2010: Modifider Ryosuke Takeuchi
 *           : Modified Device Interface to ITRON-SIL interface and Changed
 *             AT91SAM7.h to SIL defined include file.
 *           : Modified i2c_init argument for ITRON ATT_INI static API.
 *           : Modified Interrupt disable in i2c_init fuction for ITRON ATT_INI API.
 * ??/??/2009: Modifier Andy Shaw
 *           : Added PIO IED and IDR control in timer interrupt.
 */
/*
 * Υե tecsgen ˤƥץ졼ȤȤƼưޤ
 * ΥեԽƻѤ뤳ȤտޤƤޤ
 * tecsgen κƼ¹Ԥˤ񤭤Ƥޤᡢ̾
 *   gen/tI2C_template.c => src/tI2C.c
 * Τ褦̾, եѹƤ齤ޤ
 */
/* #[<PREAMBLE>]#
 * #[<...>]#  #[</...>]# ǰϤޤ줿ȤԽʤǤ
 * tecsmerge ˤޡ˻Ѥޤ
 *
 * °ޥ #_CAAM_#
 * portNumber       uint8_t          ATTR_portNumber 
 * TimerBase        uint32_t         ATTR_TimerBase  
 * deviceAddress    uint8_t*         VAR_deviceAddress
 * partialTransaction I2CPT*           VAR_partialTransaction
 * currentPT        I2CPT*           VAR_currentPT   
 * state            uint32_t         VAR_state       
 * nbits            uint32_t         VAR_nbits       
 * ackSlot          bool_t           VAR_ackSlot     
 * ackSlotPending   bool_t           VAR_ackSlotPending
 * data             uint8_t*         VAR_data        
 * transmitting     bool_t           VAR_transmitting
 *
 * #[</PREAMBLE>]# */

#include "tI2C_tecsgen.h"
#include <sil.h>
#include "target_at91sam7s.h"
#include <string.h>

typedef enum {
    I2C_UNINITIALISED = 0,
    I2C_IDLE,
    I2C_BEGIN,
    I2C_RESTART1,
    I2C_START1,
    I2C_START2,
    I2C_START3,
    I2C_START_RECLOCK1,
    I2C_LOW0,
    I2C_LOW1,
    I2C_HIGH0,
    I2C_HIGH1,
    I2C_STOP0,
    I2C_STOP1,
    I2C_STOP2,
    I2C_STOP3,
} i2c_port_state;

static const uint32_t SCL[4] = { (1 << 23), (1 << 28), (1 << 29), (1 << 30) };
static const uint32_t SDA[4] = { (1 << 18), (1 << 19), (1 << 20), (1 << 2) };

#ifndef E_OK
#define	E_OK	0		/* success */
#define	E_ID	(-18)	/* illegal ID */
#endif

static inline void
i2cDisable(CELLCB *p_cellcb)
{
    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_ODR), (SCL[ATTR_portNumber] | SDA[ATTR_portNumber]));
}

/* ؿ #_TEPF_# */
/* #[<ENTRY_PORT>]# eI2C
 * entry port: eI2C
 * signature:  sI2C
 * context:    task
 * #[</ENTRY_PORT>]# */

/* #[<ENTRY_FUNC>]# eI2C_enable
 * name:         eI2C_enable
 * global_name:  tI2C_eI2C_enable
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
void
eI2C_enable(CELLIDX idx)
{
    CELLCB	*p_cellcb;
    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }

    VAR_state = I2C_IDLE;

    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_SODR), (SCL[ATTR_portNumber] | SDA[ATTR_portNumber]));
    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_OER), (SCL[ATTR_portNumber] | SDA[ATTR_portNumber]));
    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_MDER), SCL[ATTR_portNumber]);
    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_PUDR), SDA[ATTR_portNumber]);
    sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_PUER), SCL[ATTR_portNumber]);
}

/* #[<ENTRY_FUNC>]# eI2C_disable
 * name:         eI2C_disable
 * global_name:  tI2C_eI2C_disable
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
void
eI2C_disable(CELLIDX idx)
{
    CELLCB	*p_cellcb;
    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }

    i2cDisable(p_cellcb);
}

/* #[<ENTRY_FUNC>]# eI2C_isBusy
 * name:         eI2C_isBusy
 * global_name:  tI2C_eI2C_isBusy
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
bool_t
eI2C_isBusy(CELLIDX idx)
{
    CELLCB	*p_cellcb;
    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }

    return (VAR_state > I2C_IDLE);
}

/* #[<ENTRY_FUNC>]# eI2C_startTransaction
 * name:         eI2C_startTransaction
 * global_name:  tI2C_eI2C_startTransaction
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
bool_t
eI2C_startTransaction(CELLIDX idx, uint32_t address, int32_t internalAddress, int32_t internalAddressBytes, uint8_t* data, uint32_t dataSize, int32_t direction)
{
    CELLCB	*p_cellcb;
    I2CPT *pt;

    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }

    pt = VAR_partialTransaction;
    VAR_currentPT = pt;

    memset(pt, 0, sizeof(VAR_partialTransaction));

    if(internalAddressBytes > 0){
        VAR_deviceAddress[0] = (address << 1);
        VAR_deviceAddress[1] = internalAddress;

        pt->start = 1;
        pt->stop = (direction ? 0 : 1); // readΤȤstop
        pt->tx = 1;
        pt->data = VAR_deviceAddress;
        pt->size = 2;

        pt++;
    }

    if(internalAddressBytes == 0 || !direction){  
        pt->start = (internalAddressBytes > 0) ? 0 : 1;
        pt->restart = !pt->start;
        pt->stop = 0;
        pt->tx = 1;
        VAR_deviceAddress[2] = (address << 1) | (direction ? 0 : 1);
        pt->data = &VAR_deviceAddress[2];
        pt->size = 1;
        
        pt++;
    }

    pt->start = 0;
    pt->stop = 1;
    pt->tx = (direction ? 1 : 0);
    pt->data = data;
    pt->size = dataSize;
    pt->lastPT = 1;

    // Start the transaction
    VAR_state = I2C_BEGIN;

    return true;
}

/* #[<ENTRY_FUNC>]# eI2C_initialize
 * name:         eI2C_initialize
 * global_name:  tI2C_eI2C_initialize
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
void
eI2C_initialize(CELLIDX idx)
{
    CELLCB	*p_cellcb;
    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }

    VAR_state = I2C_IDLE;
    i2cDisable(p_cellcb);
}

/* #[<ENTRY_FUNC>]# eI2C_interruptBody
 * name:         eI2C_interruptBody
 * global_name:  tI2C_eI2C_interruptBody
 * oneway:       false
 * #[</ENTRY_FUNC>]# */
void
eI2C_interruptBody(CELLIDX idx, uint32_t inputs)
{
    CELLCB	*p_cellcb;
    uint32_t codr = 0;
    uint32_t sodr = 0;
    uint32_t oer = 0;
    uint32_t odr = 0;

    if (VALID_IDX(idx)) {
        p_cellcb = GET_CELLCB(idx);
    }


    switch (VAR_state) {
      case I2C_IDLE: // ʤˤ⤷ʤ
        break;
      case I2C_BEGIN: // ̿γ
        // data, clkϵ
        oer |= SDA[ATTR_portNumber];
        oer |= SCL[ATTR_portNumber];
      
        if(VAR_currentPT && VAR_currentPT->size){
            VAR_data = VAR_currentPT->data;       // Хåե
            VAR_nbits = VAR_currentPT->size * 8;  // Хåե
            VAR_transmitting = VAR_currentPT->tx; // write
            VAR_ackSlot = 0;                      // ack
            VAR_ackSlotPending = 0;               // ackԤ
            
            if(!VAR_transmitting){ // readΤȤ
                *(VAR_data) = 0;
            }

            if(VAR_currentPT->restart){ // ϤʤȤ
                // data, clk1ˤ֡
                sodr |= SCL[ATTR_portNumber];
                sodr |= SDA[ATTR_portNumber];
                VAR_state = I2C_RESTART1;
            } else if(VAR_currentPT->start){ // ƳϤȤ
                // data = 1
                sodr |= SDA[ATTR_portNumber];
                VAR_state = I2C_START1;
            } else {
                codr |= SCL[ATTR_portNumber];
                VAR_state = I2C_LOW0;
            }
        } else { // error
            VAR_state = I2C_IDLE;
        }
        break;
      case I2C_RESTART1: // data = 1, clk = 1
        // clk = 0
        codr |= SCL[ATTR_portNumber];
        VAR_state = I2C_START1;
        break;
      case I2C_START1: // data = 1, clk = 0
        // SDA high, take SCL high
        // clk = 1
        sodr |= SCL[ATTR_portNumber];
        VAR_state = I2C_START2;
        break;
      case I2C_START2: // data = 1, clk = 1
        if(inputs & SDA[ATTR_portNumber]){ // data = 1ΤȤ
            // data = 0  start
            codr |= SDA[ATTR_portNumber];
            VAR_state = I2C_START3;
        } else { // error  ľ
            // clk = 0
            codr |= SCL[ATTR_portNumber];
            VAR_state = I2C_START_RECLOCK1;
        }
        break;
      case I2C_START_RECLOCK1:
        // clk = 0
        codr |= SCL[ATTR_portNumber];
        VAR_state = I2C_START1;
        break;
      case I2C_START3: // data = 0, clk = 1ϡ
        // clk = 0  åΩ
        codr |= SCL[ATTR_portNumber];
        VAR_state = I2C_LOW0;
        break;
      case I2C_LOW0: // clk = 0, data = 1ǡ
        if(VAR_ackSlotPending){ // ackԤΤȤ
            VAR_ackSlot = 1; // ack
            VAR_ackSlotPending = 0;
        } else { // ackԤǤʤ
            VAR_ackSlot = 0;
        }

        if(VAR_nbits || VAR_ackSlot){ //  or ack
            if(VAR_ackSlot) { // ack
                
                if(VAR_transmitting){ // ¦ΤȤ
                    // dataؤν߶ػ
                    odr |= SDA[ATTR_portNumber];
                } else { // ¦ΤȤ
                    // data = 0 ack
                    oer  |= SDA[ATTR_portNumber];
                    codr |= SDA[ATTR_portNumber];
                }
            } else if(!VAR_transmitting){ // ackǤʤ¦
                // dataؤν߶ػ
                odr |= SDA[ATTR_portNumber];
            } else { // ackǤʤ¦
                // dataؤνߵ
                oer |= SDA[ATTR_portNumber];
                // ǡ1bit
                VAR_nbits--;
                if(((*(VAR_data)) >> (VAR_nbits & 7)) & 0x01){ // ǡ1
                    // data = 1
                    sodr |= SDA[ATTR_portNumber];
                } else { // ǡ0
                    // data = 0
                    codr |= SDA[ATTR_portNumber];
                }
                
                if((VAR_nbits & 7) == 0){ // 8bitä  ackԤ
                     // Хåե򹹿Ƽ8bitΥǡ
                    VAR_data++;
                    if(VAR_nbits || VAR_transmitting){
                        VAR_ackSlotPending = 1; // ackԤ
                    }
                }
            }
            VAR_state = I2C_LOW1;
            
        } else if(VAR_currentPT->stop){ // ä
            VAR_state = I2C_STOP0;
        } else { // Υǡ򳫻Ϥ
            VAR_currentPT++;
            // data = 1
            sodr |= SDA[ATTR_portNumber];
            VAR_state = I2C_BEGIN;
        }
        break;
      case I2C_LOW1: // clk = 0, data = ǡ
        // clk = 1  åΩ夬
        sodr |= SCL[ATTR_portNumber];
        VAR_state = I2C_HIGH0;
        break;
      case I2C_HIGH0: // clk = 1, data = ǡ
        if(inputs & SCL[ATTR_portNumber]){ // clk = 1ˤʤޤԤ
            VAR_state = I2C_HIGH1;
        }
        break;
      case I2C_HIGH1: // clk = 1, data = ǡ
        if(VAR_transmitting && VAR_ackSlot){ // ¦ and ackԤ
            if(inputs & SDA[ATTR_portNumber]){ // error  ack data = 0
                // stop
                VAR_state = I2C_STOP0;
            } else { // ack
                // Υåȯ
                VAR_state = I2C_LOW0;
            }
            
            // clk = 0
            codr |= SCL[ATTR_portNumber];
        } else { // ¦ or ackԤǤʤ        
            if(!VAR_transmitting && !VAR_ackSlot){ // ¦ and ackԤǤʤ
                // ǡ1bitɤ߹
                VAR_nbits--;
                if(inputs & SDA[ATTR_portNumber]){ // Ϥ1
                    *(VAR_data) |= (1 << (VAR_nbits & 7));
                }
                
                if(VAR_nbits && ((VAR_nbits & 7) == 0)){ // ǡ8bit
                    // ΥХåե
                    VAR_data++;
                    VAR_ackSlotPending = 1; // ackԤ
                    *(VAR_data) = 0;
                }
            }

            // clk = 0
            codr |= SCL[ATTR_portNumber];
            VAR_state = I2C_LOW0;
        }
        break;
      case I2C_STOP0: // clk = 0stopϡ
        // data = 0
        oer  |= SDA[ATTR_portNumber];
        codr |= SDA[ATTR_portNumber];
        VAR_state = I2C_STOP1;
        break;
      case I2C_STOP1: // data = 0, clk = 0
        // clk = 1
        sodr |= SCL[ATTR_portNumber];
        VAR_state = I2C_STOP2;
        break;
      case I2C_STOP2: // data = 0, clk = 1
        // data = 1  stop
        sodr |= SDA[ATTR_portNumber];
        VAR_state = I2C_STOP3;
        break;
      case I2C_STOP3: // data = 1, clk = 1stop
        if(VAR_currentPT->lastPT){ // ̿νλ
            VAR_state = I2C_IDLE; // Ե֤
        } else { // ̿
            VAR_currentPT++;
            VAR_state = I2C_BEGIN;
        }
        break;
      case I2C_UNINITIALISED:	// Uninitialised
        break;
      default:
        break;
    } // switch


    /*
     *  ºݤ˥쥸˽񤭹
     *  ǽ񤭹쥸ϡ
     *  1񤭹Τͭǡ0񤭹̵Ǥ롥
     *  Ĥޤꡤ񤭹ߤӥåȤ1Ȥ񤭹ߤʤӥåȤ
     *  0ȤФ褤
     */
    if(codr){
        sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_CODR), codr);
    }
    if(sodr){
        sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_SODR), sodr);
    }
    if(oer){
        sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_OER), oer);
    }
    if(odr){
        sil_wrw_mem((void *)(TADR_PIO_BASE+TOFF_PIO_ODR), odr);
    }
}

/* #[<POSTAMBLE>]#
 *   겼ؿ񤭤ޤ
 * #[</POSTAMBLE>]#*/
